try:
    import time

    from pycrc import algorithms
    import anycrc
    import crcmod.predefined
    import fastcrc
    import crc_ct #renamed from crc
    import libscrc
    import crcengine
    import crccheck
    import crc

    N_C_EXT = 10000
    N_PYTHON = 100
    LENGTH = 65536

    class Benchmark:
        def __init__(self, module, n, relative_to=None):
            self.module = module
            self.n = n
            self.average = None
            self.relative_to = relative_to
            self.start = time.perf_counter()
            self.ran_once = False

        def __repr__(self):
            if self.ran_once:
                return f'{self.module}\nTime: {self.average * self.n:.2f}s\nSpeed: {self.get_speed():.2f} MB/s\nRelative: {self.get_relative():.2f}'
            else:
                raise Exception("Benchmark was not completed")

        def stop(self):
            t = time.perf_counter()
            self.average = (t - self.start) / self.n
            self.ran_once = True
            self.start = t

        def get_speed(self):
            return LENGTH / self.average / (10 ** 6)

        def get_relative(self):
            if self.relative_to is None:
                return 1
            else:
                return self.average / self.relative_to.average

    benchmarks = []

    data = bytes(i & 0xff for i in range(LENGTH))

    print(f'\nN = {N_C_EXT}\n')

    #anycrc
    anycrc_benchmark = Benchmark('anycrc', N_C_EXT)

    model = anycrc.Model('CRC32')
    for i in range(anycrc_benchmark.n):
        model.calc(data)

    anycrc_benchmark.stop()
    benchmarks.append(anycrc_benchmark)

    print(anycrc_benchmark)
    print()

    #fastcrc
    benchmark = Benchmark('fastcrc', N_C_EXT, relative_to=anycrc_benchmark)

    for i in range(benchmark.n):
        fastcrc.crc32.iso_hdlc(data)

    benchmark.stop()
    benchmarks.append(benchmark)

    print(benchmark)
    print()

    #crcmod
    benchmark = Benchmark('crcmod-plus', N_C_EXT, relative_to=anycrc_benchmark)

    calc = crcmod.predefined.mkPredefinedCrcFun('crc-32')
    for i in range(benchmark.n):
        calc(data)

    benchmark.stop()
    benchmarks.append(benchmark)

    print(benchmark)
    print()

    #crc-ct
    benchmark = Benchmark('crc-ct', N_C_EXT, relative_to=anycrc_benchmark)

    crc_model = crc_ct.predefined_model_by_name(b'CRC-32')
    for i in range(benchmark.n):
        crc_result = crc_ct.init(crc_model)
        crc_result = crc_ct.update(crc_model, data, len(data), crc_result)
        crc_ct.final(crc_model, crc_result)

    benchmark.stop()
    benchmarks.append(benchmark)

    print(benchmark)
    print()

    #libscrc
    benchmark = Benchmark('libscrc', N_C_EXT, relative_to=anycrc_benchmark)

    for i in range(benchmark.n):
        libscrc.crc32(data)

    benchmark.stop()
    benchmarks.append(benchmark)

    print(benchmark)
    print()

    #use smaller N for slower pure python libraries
    print(f'N = {N_PYTHON}\n')

    #crcengine
    benchmark = Benchmark('crcengine', N_PYTHON, relative_to=anycrc_benchmark)

    crc_algorithm = crcengine.new('crc32')
    for i in range(benchmark.n):
        crc_algorithm(data)

    benchmark.stop()
    benchmarks.append(benchmark)

    print(benchmark)
    print()

    #pycrc
    benchmark = Benchmark('pycrc', N_PYTHON, relative_to=anycrc_benchmark)

    CRC = algorithms.Crc(width=32, poly=0x4c11db7, reflect_in=True, xor_in=0xffffffff, reflect_out=True, xor_out=0xffffffff)
    for i in range(benchmark.n):
        CRC.table_driven(data)

    benchmark.stop()
    benchmarks.append(benchmark)

    print(benchmark)
    print()

    #crccheck
    benchmark = Benchmark('crccheck', N_PYTHON, relative_to=anycrc_benchmark)

    for i in range(benchmark.n):
        crcinst = crccheck.crc.Crc32()
        crcinst.process(data)
        crcinst.final()

    benchmark.stop()
    benchmarks.append(benchmark)

    print(benchmark)
    print()

    #crc
    benchmark = Benchmark('crc', N_PYTHON, relative_to=anycrc_benchmark)

    calculator = crc.Calculator(crc.Crc32.CRC32, optimized=True)
    for i in range(benchmark.n):
        calculator.checksum(data)

    benchmark.stop()
    benchmarks.append(benchmark)

    print(benchmark)
    print()

    benchmarks.sort(key=lambda benchmark: benchmark.get_speed(), reverse=True)

    file_name = 'benchmark_results.txt'

    with open(file_name, 'w') as file:
        file.write('| Module | Speed (MB/s) | Relative |\n')
        file.write('|---|:-:|:-:|\n')

        for benchmark in benchmarks:
            file.write(f'| {benchmark.module} | {benchmark.get_speed():.2f} | x{benchmark.get_relative():.2f} |\n')

    print(f'Results saved to "{file_name}"')

except KeyboardInterrupt:
    pass


# This script creates a virtual environment, installs all dependancies, and runs the benchmark
try:
    import os
    import shutil
    import subprocess
    import sys

    packages_path = '.venv/Lib/site-packages'

    if os.path.isdir(packages_path):
        for file in os.listdir(packages_path):
            if file == 'crc' or file.startswith('crc-') or file == 'crc_ct' or file.startswith('crc_ct-'):
                shutil.rmtree(os.path.join(packages_path, file))

    modules = ['fastcrc', 'crcmod-plus', 'crc-ct', 'libscrc', 'crcengine', 'pycrc', 'crccheck']

    subprocess.run(['python', '-m' 'venv', '.venv'])
    subprocess.run(['.venv/Scripts/pip', 'install'] + modules)

    # This is necessary due to a naming conflict between crc and crc-ct
    shutil.move(os.path.join(packages_path, 'crc'), os.path.join(packages_path, 'crc_ct'))
    subprocess.run(['.venv/Scripts/pip', 'install', 'crc'])
    
    subprocess.run(['.venv/Scripts/pip', 'install', '..'])

    subprocess.run(['.venv/Scripts/python', 'bench.py'])

except KeyboardInterrupt:
    pass


# Download the latest models from the CRC RevEng website

from urllib.request import urlopen
from bs4 import BeautifulSoup

url = 'https://reveng.sourceforge.io/crc-catalogue/all.htm'

with urlopen(url) as response:
    page = response.read()

soup = BeautifulSoup(page, 'html.parser')
tags = soup.select('h3 + p > code, h3 + p + ul > li > a > strong')

crcs = {}
aliases = {}

current_crc = ''
current_width = 0

for tag in tags:
    if tag.name == 'code':
        name = tag.text.partition('name="')[2].rstrip('"')
        name = name.replace('-', '', 1).replace('/', '-')

        params = tag.text.partition('  residue=')[0]
        params = params.replace('  ', ', ').replace('true', 'True').replace('false', 'False')

        width_start = tag.text.find('width=') + len('width=')
        width_end = tag.text.find('  ', width_start)
        width = int(params[width_start:width_end])

        current_crc = name
        current_width = width

        if width <= 64:
            crcs[name] = params

    elif tag.name == 'strong' and current_width <= 64:
        name = tag.text.replace('/', '-')
        if name.startswith('CRC-') and name[4].isnumeric():
            name = name.replace('-', '', 1)

        aliases[name] = current_crc

with open('models.py', 'w') as file:
    file.write("""# Taken from Greg Cook's CRC catalogue: https://reveng.sourceforge.io/crc-catalogue/all.htm

from collections import namedtuple

model = namedtuple('model', ['width', 'poly', 'init', 'refin', 'refout', 'xorout', 'check'], defaults=(0,))

""")

    file.write('models = {\n')

    for model, params in crcs.items():
        file.write(f"    '{model}': model({params}),\n")

    file.write('}\n\n')
    file.write('aliases = {\n')

    for alias, model in aliases.items():
        file.write(f"    '{alias}': '{model}',\n")

    file.write('}')


from setuptools import setup, Extension

setup(
    ext_modules = [
        Extension(
            name='anycrc.crc',
            sources=['src/anycrc/crc.pyx',
                     'lib/crcany/model.c',
                     'lib/crcany/crc.c']
        )
    ]
)


# Copyright (c) 2024-2025 Hussain Al Marzooq

from bitarray import bitarray, frozenbitarray
from .models import models, aliases

cdef extern from '../../lib/crcany/model.h':
    ctypedef unsigned long long word_t
    cdef const unsigned short WORDBITS

    ctypedef struct model_t:
        unsigned short width
        char ref, rev
        word_t poly, init, xorout
        word_t *table
        word_t *table_comb

    cdef model_t get_model(unsigned short width, word_t poly, word_t init, char refin, char refout, word_t xorout)
    cdef char init_model(model_t *model)
    cdef void free_model(model_t *model)

cdef extern from '../../lib/crcany/crc.h':
    cdef void crc_table_bytewise(model_t *model)
    cdef word_t crc_bytewise(model_t *model, word_t crc, const void *dat, size_t len)

    cdef void crc_table_slice16(model_t *model)
    cdef word_t crc_slice16(model_t *model, word_t crc, const void *dat, size_t len)

    cdef void crc_table_combine(model_t *model)
    word_t crc_combine(model_t *model, word_t crc1, word_t crc2, size_t len2)

cache = {}

cdef class _Crc:
    cdef model_t model

    def __cinit__(self, width, poly, init, refin, refout, xorout, check=None):
        self.model = get_model(width, poly, init, refin, refout, xorout)
        cdef char error_code = init_model(&self.model)

        if error_code == 1:
            raise MemoryError('Out of memory error')

        crc_table_bytewise(&self.model)
        crc_table_slice16(&self.model)
        crc_table_combine(&self.model)

    def __dealloc__(self):
        free_model(&self.model);

    def calc(self, data, init=None):
        if isinstance(data, bitarray) or isinstance(data, frozenbitarray):
            raise TypeError('Bitarray objects are not allowed, use calc_bits() instead')

        if isinstance(data, str):
            data = (<unicode> data).encode('utf-8')

        if init is None:
            init = self.model.init

        if len(data) == 0:
            return init

        cdef const unsigned char[:] view = data
        return crc_slice16(&self.model, init, &view[0], len(view) * 8)

    def calc_bits(self, data, init=None):
        if not isinstance(data, bitarray) and not isinstance(data, frozenbitarray):
            raise TypeError('Expected a bitarray object')

        if self.model.ref and data.endian != 'little':
            raise ValueError('A little endian bitarray object is expected for reflected CRCs')

        if not self.model.ref and data.endian != 'big':
            raise ValueError('A big endian bitarray object is expected for non-reflected CRCs')

        if init is None:
            init = self.model.init

        if len(data) == 0:
            return init

        cdef const unsigned char[:] view = data
        return crc_slice16(&self.model, init, &view[0], len(data))

    def combine(self, crc1, crc2, length):
        return crc_combine(&self.model, crc1, crc2, length * 8)

    def combine_bits(self, crc1, crc2, length):
        return crc_combine(&self.model, crc1, crc2, length)

    #byte-by-byte (for testing)
    def _calc_b(self, data):
        cdef const unsigned char[:] view = data
        return crc_bytewise(&self.model, self.model.init, &view[0], len(view) * 8)

def CRC(width=None, poly=None, init=None, refin=None, refout=None, xorout=None, check=None):
    names = ('width', 'poly', 'init', 'refin', 'refout', 'xorout')
    values = (width, poly, init, refin, refout, xorout)

    for name, value in zip(names, values):
        if value is None:
            raise ValueError(f'{name} value is not provided')

    if width > WORDBITS:
        raise ValueError(f'width is larger than {WORDBITS} bits')

    if values in cache:
        return cache[values]
    else:
        crc = _Crc(*values)
        cache[values] = crc
        return crc

def Model(name):
    if name in models:
        return CRC(*models[name])
    elif name in aliases:
        return CRC(*models[aliases[name]])
    else:
        raise ValueError('CRC model not found')


from bitarray import bitarray
import anycrc

#The standard test data b'123456789' isn't good enough when testing slice-by-16 because the data length is < 16
test_data = b'123456789'
test_data2 = b'abcdefghijklmnopqrstuvwxyz'

for name, model in anycrc.models.items():
    print(name)
    print(anycrc.str_model(model))

    crc = anycrc.Model(name)
    crc2 = anycrc.Model(name)

    #check to see if the result from the byte-by-byte algorithm matches the check value
    value = crc._calc_b(test_data)
    check = model.check
    print('byte-by-byte:  {} {}'.format(anycrc.get_hex(value, model.width), anycrc.get_hex(check, model.width)))
    #assert value == check

    #check to see if the result from the slice-by-16 algorithm matches the byte-by-byte value
    value = crc.calc(test_data2)
    value2 = crc2._calc_b(test_data2)

    print('slice-by-16:   {} {}'.format(anycrc.get_hex(value, model.width), anycrc.get_hex(value2, model.width)))
    #assert value == value2

    #process the data one byte at a time
    value = None
    for c in test_data:
        value = crc.calc(c.to_bytes(1, 'little'), value)

    print('update:        {} {}'.format(anycrc.get_hex(value, model.width), anycrc.get_hex(check, model.width)))
    #assert value == check

    #with length in bits
    if model.refin:
        bit_data = bitarray(endian='little')
    else:
        bit_data = bitarray()

    bit_data.frombytes(test_data2)
    value = crc.calc_bits(bit_data[:100])
    value = crc.calc_bits(bit_data[100:], value)
    value2 = crc2._calc_b(test_data2)

    print('bits:          {} {}'.format(anycrc.get_hex(value, model.width), anycrc.get_hex(value2, model.width)))
    #assert value == value2

    #combine
    value = crc.calc(b'12345')
    value2 = crc.calc(b'6789')
    value3 = crc.combine(value, value2, len(b'6789'))
    print('combine:       {} {}'.format(anycrc.get_hex(value3, model.width), anycrc.get_hex(check, model.width)))
    #assert value3 == check

    #combine bits
    if model.refin:
        bit_data = bitarray(endian='little')
    else:
        bit_data = bitarray()

    bit_data.frombytes(test_data)
    value = crc.calc_bits(bit_data[:36])
    value2 = crc.calc_bits(bit_data[36:])
    value3 = crc.combine_bits(value, value2, len(bit_data[36:]))
    print('combine bits:  {} {}'.format(anycrc.get_hex(value3, model.width), anycrc.get_hex(check, model.width)))
    #assert value3 == check

    print()

#check that all of the aliases are valid
for alias, name in anycrc.aliases.items():
    print(alias, end=' ')
    #assert name in anycrc.models
    print('OK')

print()